# SPDX-License-Identifier: AGPL-3.0-only

import itertools
import math
import os
import sys
import shlex
import subprocess as sub

import cv2
import numpy as np
import supervision as sv
import torch
import torchvision
import tqdm

import groundingdino.util.inference as gd
# import segment_anything_hq as samhq # Noisy

from ezsam.lib.date import now
from ezsam.lib.file import InputMode, get_input_mode
from ezsam.formats import OutputImageFormat, OutputVideoCodec, get_video_fmt_from_codec


def process_file(
  src: str,
  prompts: list[str],
  neg_prompts: list[str],
  box_threshold: float,
  text_threshold: float,
  nms_threshold: float,
  sam_predictor,  #: samhq.SamPredictor,
  grounding_dino_model: gd.Model,
  img_fmt: OutputImageFormat,
  codec: OutputVideoCodec,
  num_test_frames: int,
  output_suffix: str,
  output_dir: str,
  debug: bool,
  cleanup: bool,
) -> None:
  input_mode = get_input_mode(src)
  # Determine output extension: preserve for images in debug mode, else use formats that support transparency.
  input_filename, input_ext = os.path.splitext(os.path.basename(src))
  ext = input_ext
  if input_mode == InputMode.image and not debug:
    ext = '.' + img_fmt
  elif input_mode == InputMode.video:
    ext = '.' + get_video_fmt_from_codec(codec)
  out = output_dir + '/' + input_filename + output_suffix + ext
  print(f'{now()}: Processing file {src} to {out} ...')
  process_image_args = {
    'prompts': prompts,
    'neg_prompts': neg_prompts,
    'box_threshold': box_threshold,
    'text_threshold': text_threshold,
    'nms_threshold': nms_threshold,
    'sam_predictor': sam_predictor,
    'grounding_dino_model': grounding_dino_model,
    'debug': debug,
  }
  print(f'Process image args: {process_image_args}')

  if input_mode == InputMode.image:
    # Load image, discarding any alpha channel information if present
    image: np.ndarray = cv2.imread(src)  # note: default method cv2.IMREAD_COLOR, format BGR
    # Version of image with alpha information if present. Note that BGR is default colour mode using OpenCV library (cv2).
    image_unchanged: np.ndarray = cv2.imread(src, cv2.IMREAD_UNCHANGED)
    processed_image = process_image(image=image, image_unchanged=image_unchanged, **process_image_args)
    cv2.imwrite(out, processed_image)

  elif input_mode == InputMode.video:
    print(f'Using extension / codec: {ext} / {codec} ...')

    # Process all input frames to temporary image files
    tmp_files = []
    video_frames_generator = sv.get_video_frames_generator(source_path=src)
    video_info = sv.VideoInfo.from_video_path(video_path=src)
    fps = video_info.fps
    (w, h) = video_info.resolution_wh
    i = 0
    # I.e. 10 frames => 1 digit, 0..9. 11 frames => 2 digits, 00..10.
    num_digits = int(math.log10(video_info.total_frames - 1)) + 1
    if num_test_frames is None:
      total = video_info.total_frames
      frame_gen = video_frames_generator
    else:
      total = num_test_frames
      frame_gen = itertools.islice(video_frames_generator, total)
    for frame in tqdm.tqdm(frame_gen, total=total):
      # Pad counter to num_digits
      i_pad = str(i).zfill(num_digits)
      tmp = f'{output_dir}/{input_filename}.{i_pad}.tmp.{img_fmt}'
      processed_image = process_image(image=frame, image_unchanged=None, **process_image_args)
      i += 1
      print(f'Writing frame {i} to {tmp} ...')
      cv2.imwrite(tmp, processed_image)
      tmp_files.append(tmp)

    # Join temporary processed images into video using either FFmpeg or ImageMagick's convert
    if codec == OutputVideoCodec.gif:
      tmp_img_naming = f'{output_dir}/{input_filename}.*.tmp.{img_fmt}'
      cmd_in = ''
    else:
      tmp_img_naming = f'{output_dir}/{input_filename}.%{num_digits}d.tmp.{img_fmt}'
      cmd_in = f'ffmpeg -y -framerate {fps} -i {tmp_img_naming}'
    # ref: https://stackoverflow.com/a/75461590
    # Note: Software support is iffy for all but gif.
    # Chrome can display alpha for vp9+webm.
    # mpv works for the rest.
    cmd_out = ''
    if codec == OutputVideoCodec.prores:
      cmd_out = f'-c:v prores -pix_fmt yuva444p10le {out}'
    elif codec == OutputVideoCodec.vp9:
      cmd_out = f'-c:v libvpx-vp9 -pix_fmt yuva420p {out}'
    elif codec == OutputVideoCodec.ffv1:
      cmd_out = f'-c:v ffv1 -pix_fmt yuva420p {out}'
    elif codec == OutputVideoCodec.apng:
      cmd_out = f'-c:v apng -pix_fmt rgba {out}'
    elif codec == OutputVideoCodec.gif:
      delay = get_delay_from_fps(fps)
      cmd_out = f'convert -resize {w}x{h} -delay {delay} -dispose Background -loop 0 "{tmp_img_naming}" {out}'
    else:
      raise ValueError(f'Invalid codec: {codec}')
    cmd = cmd_in + ' ' + cmd_out
    print(f'Joining video frames via command: {cmd} ...')
    cmd_args = shlex.split(cmd)
    sub.run(cmd_args)

    if cleanup:
      for tmp in tmp_files:
        try:
          print(f'Deleting temp file: {tmp} ...')
          os.remove(tmp)
        except Exception as err:
          print(f'Error deleting temporary image file {tmp}')
          print(f'{err}')


def get_delay_from_fps(fps):
  # Get centiseconds delay from frames per second, used as ImageMagick's delay parameter
  f = fps if (fps is not None and fps != 0) else 1
  return 100.0 / f


def get_video_codec(src: str):
  codec = None
  try:
    video_capture = cv2.VideoCapture(src)
    codec_code = video_capture.get(cv2.CAP_PROP_FOURCC)
    # Convert OpenCV codec code which is a float to 4 character string
    codec = int(codec_code).to_bytes(4, byteorder=sys.byteorder).decode()
  except Exception:
    print(f'Error getting codec for video {src}')
  return codec


def process_image(
  image: np.ndarray,
  # Masks are generated using alpha-stripped version of image, but output image is original with added masks
  image_unchanged: np.ndarray | None,
  prompts: list[str],
  neg_prompts: list[str],
  box_threshold: float,
  text_threshold: float,
  nms_threshold: float,
  sam_predictor,  #: samhq.SamPredictor,
  grounding_dino_model: gd.Model,
  debug: bool,
) -> np.ndarray:
  print('Processing image...')
  if debug:
    print(f'{now()} Joining prompts for debug mode ...')
    prompts = prompts + (neg_prompts or [])
  else:
    print(f'{now()} Handling positive prompts...')
  detections = detections_for_image(
    grounding_dino_model=grounding_dino_model,
    image=image,
    prompts=prompts,
    box_threshold=box_threshold,
    text_threshold=text_threshold,
    nms_threshold=nms_threshold,
    sam_predictor=sam_predictor,
  )
  if detections is None:
    if debug:
      print('Returning original image ...')
      return image
    else:
      print('Returning empty image ...')
      # Create a new image with dimensions of old image plus an alpha channel, and then zero out everything
      processed_image = cv2.cvtColor(image, cv2.COLOR_BGR2BGRA)
      processed_image[:, :, :] = 0
      return processed_image

  neg_detections: sv.Detections | None
  has_neg_prompts = neg_prompts and len(neg_prompts) > 0
  if not debug and has_neg_prompts:
    print(f'{now()} Handling negative prompts...')
    neg_detections = detections_for_image(
      grounding_dino_model=grounding_dino_model,
      image=image,
      prompts=neg_prompts,
      box_threshold=box_threshold,
      text_threshold=text_threshold,
      nms_threshold=nms_threshold,
      sam_predictor=sam_predictor,
    )

  if debug:
    print(f'{now()} Annotating output image ...')
    # Annotate image with SAM segment masks and GroundingDINO object detection boxes.
    # Note: Should set ColorLookup.INDEX when annotating for SAM.
    # ref: https://github.com/roboflow/notebooks/blob/main/notebooks/how-to-segment-anything-with-sam.ipynb
    # ref: https://supervision.roboflow.com/annotators/
    mask_annotator = sv.MaskAnnotator(color_lookup=sv.ColorLookup.INDEX)
    box_corner_annotator = sv.BoxCornerAnnotator(color_lookup=sv.ColorLookup.INDEX)
    label_annotator = sv.LabelAnnotator(text_position=sv.Position.CENTER_OF_MASS, color_lookup=sv.ColorLookup.INDEX)
    labels = get_labels(prompts, detections)
    processed_image = mask_annotator.annotate(scene=image.copy(), detections=detections)
    processed_image = box_corner_annotator.annotate(scene=processed_image, detections=detections)
    processed_image = label_annotator.annotate(scene=processed_image, detections=detections, labels=labels)
  else:
    print(f'{now()} Filtering output image ...')
    # Filter image using masks...
    # First join all masks together; reduce on first axis, since that's the mask number in detections.mask.
    # Note: detection.mask is array of n masks * H pixels * W pixels, with each pixel True or False.
    pos_supermask: np.ndarray = np.logical_or.reduce(detections.mask, axis=0)
    if has_neg_prompts and neg_detections:
      # Negative mask is just flipped
      neg_supermask: np.ndarray = ~np.logical_or.reduce(neg_detections.mask, axis=0)
      # Joint removes negative mask from positive
      supermask = np.logical_and(pos_supermask, neg_supermask)
    else:
      supermask = pos_supermask
    # We prefer basing output on original image including any alpha channel, if present
    processed_image = cv2.cvtColor(image if not is_ndarray(image_unchanged) else image_unchanged, cv2.COLOR_BGR2BGRA)
    # Apply mask to image's alpha channel
    processed_image[:, :, 3] = np.multiply(processed_image[:, :, 3], supermask)
  return processed_image


def is_ndarray(array: np.ndarray):
  return type(array) is np.ndarray


def detections_for_image(
    grounding_dino_model: gd.Model,
    image,
    prompts: list[str],
    box_threshold: float,
    text_threshold: float,
    nms_threshold: float,
    sam_predictor,  #: samhq.SamPredictor,
  ) -> sv.Detections:

  # Detect objects
  detections: sv.Detections = grounding_dino_model.predict_with_classes(
    image=image, classes=prompts, box_threshold=box_threshold, text_threshold=text_threshold
  )

  # NMS post processing to remove lower quality boxes
  print(f'{now()} Before NMS: {len(detections.xyxy)} boxes')
  nms_idx = (
    torchvision.ops.nms(torch.from_numpy(detections.xyxy), torch.from_numpy(detections.confidence), nms_threshold)
    .numpy()
    .tolist()
  )
  detections.xyxy = detections.xyxy[nms_idx]
  detections.confidence = detections.confidence[nms_idx]
  detections.class_id = detections.class_id[nms_idx]
  num_detections = len(detections.xyxy)
  print(f'{now()} After NMS: {num_detections} boxes')

  if num_detections <= 0:
    print(f'Warning: no objects detected for prompts {prompts}')
    return None

  import segment_anything_hq as samhq

  def segment(sam_predictor: samhq.SamPredictor, image: np.ndarray, xyxy: np.ndarray) -> np.ndarray:
    # Prompt SAM with boxes for all detected objects
    sam_predictor.set_image(image, 'BGR')
    result_masks = []
    for box in xyxy:
      masks, scores, logits = sam_predictor.predict(box=box, multimask_output=True)
      index = np.argmax(scores)
      result_masks.append(masks[index])
    return np.array(result_masks)
  
  print(f'{now()} Converting object detections to segment masks ...')
  detections.mask = segment(sam_predictor=sam_predictor, image=image, xyxy=detections.xyxy)

  return detections


def get_labels(prompts: list[str], detections: sv.Detections) -> list[str]:
  if not prompts or len(prompts) <= 0:
    raise ValueError('get_labels: No prompts')
  if not detections or len(detections) <= 0:
    raise ValueError('get_labels: No detections')

  def generate_label(class_id, confidence):
    if class_id is None:
      label = 'Error'
    else:
      label = prompts[class_id] if len(prompts) > class_id else f'Class {class_id}'
    if confidence is None:
      confidence = 0
    return f'{label} {confidence:0.2f}'

  return [generate_label(class_id, confidence) for _, _, confidence, class_id, _ in detections]



from ezsam.formats import OutputImageFormat, OutputVideoCodec


DEFAULT_SAM_MODEL = 'vit_h'
DEFAULT_GROUNDING_DINO_CONFIG_PATH = './config/GroundingDINO_SwinT_OGC.py'
DEFAULT_GROUNDING_DINO_CONFIG_PATH_TMP = '.gdconf.py'
DEFAULT_OUTPUT_DIR = '.'
DEFAULT_OUTPUT_SUFFIX = '.out'
DEFAULT_BOX_THRESHOLD = 0.3
DEFAULT_TEXT_THRESHOLD = 0.3
DEFAULT_NMS_THRESHOLD = 0.8
DEFAULT_IMAGE_FORMAT = OutputImageFormat.png.value
DEFAULT_VIDEO_CODEC = OutputVideoCodec.vp9.value



import os
import shutil
import requests

from ezsam.lib.date import now


# Download a file from a given URL to the specified directory
def download(url: str, outdir: str):
  print(f'{now()}: Downloading {url} to {outdir} ...')
  if not os.path.exists(outdir):
    os.makedirs(outdir)
  file_name = os.path.basename(url)
  file_path = os.path.join(outdir, file_name)
  response = requests.get(url, stream=True)
  with open(file_path, 'wb') as file:
    shutil.copyfileobj(response.raw, file)



# Convenience function to clone a copy of config/GroundingDINO_SwinT_OGC.py 
#  for grounding.dino.inference.Model's constructor

import os

from ezsam.config.defaults import (
  DEFAULT_GROUNDING_DINO_CONFIG_PATH_TMP,
)


def create_gdconfig_tmpfile(src=DEFAULT_GROUNDING_DINO_CONFIG_PATH_TMP):
  print(f'Creating temp GroundingDINO config file at: {src} ...')
  with open(src, 'w') as f:
    f.write(str(DEFAULT_GD_CONFIG))
    f.close()
  if not os.path.isfile(src):
    raise Exception(f'Could not create temp file: {DEFAULT_GROUNDING_DINO_CONFIG_PATH_TMP}')
  return src


def cleanup_gdconfig_tmpfile(src=DEFAULT_GROUNDING_DINO_CONFIG_PATH_TMP):
  print(f'Cleaning up temp GroundingDINO config file at: {src} ...')
  if os.path.isfile(src):
    os.remove(src)


DEFAULT_GD_CONFIG = '''
batch_size = 1
modelname = 'groundingdino'
backbone = 'swin_T_224_1k'
position_embedding = 'sine'
pe_temperatureH = 20
pe_temperatureW = 20
return_interm_indices = [1, 2, 3]
backbone_freeze_keywords = None
enc_layers = 6
dec_layers = 6
pre_norm = False
dim_feedforward = 2048
hidden_dim = 256
dropout = 0.0
nheads = 8
num_queries = 900
query_dim = 4
num_patterns = 0
num_feature_levels = 4
enc_n_points = 4
dec_n_points = 4
two_stage_type = 'standard'
two_stage_bbox_embed_share = False
two_stage_class_embed_share = False
transformer_activation = 'relu'
dec_pred_bbox_embed_share = True
dn_box_noise_scale = 1.0
dn_label_noise_ratio = 0.5
dn_label_coef = 1.0
dn_bbox_coef = 1.0
embed_init_tgt = True
dn_labelbook_size = 2000
max_text_len = 256
text_encoder_type = 'bert-base-uncased'
use_text_enhancer = True
use_fusion_layer = True
use_checkpoint = True
use_transformer_ckpt = True
use_text_cross_attention = True
text_dropout = 0.0
fusion_dropout = 0.0
fusion_droppath = 0.1
sub_sentence_present = True
'''


from enum import Enum


class Model(str, Enum):
  vit_h = 'vit_h'
  vit_l = 'vit_l'
  vit_b = 'vit_b'
  hq_vit_h = 'hq_vit_h'
  hq_vit_l = 'hq_vit_l'
  hq_vit_b = 'hq_vit_b'
  hq_vit_tiny = 'hq_vit_tiny'
  gd = 'gd'


MODEL_URL = {
  Model.vit_h: 'https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth',
  Model.vit_l: 'https://dl.fbaipublicfiles.com/segment_anything/sam_vit_l_0b3195.pth',
  Model.vit_b: 'https://dl.fbaipublicfiles.com/segment_anything/sam_vit_b_01ec64.pth',
  Model.hq_vit_h: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_h.pth',
  Model.hq_vit_l: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_l.pth',
  Model.hq_vit_b: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_b.pth',
  Model.hq_vit_tiny: 'https://huggingface.co/lkeab/hq-sam/resolve/main/sam_hq_vit_tiny.pth',
  Model.gd: 'https://github.com/IDEA-Research/GroundingDINO/releases/download/v0.1.0-alpha/groundingdino_swint_ogc.pth',
}


DEFAULT_MODEL_LOCATION = {
  Model.vit_h: './models/sam_vit_h_4b8939.pth',
  Model.vit_l: './models/sam_vit_l_0b3195.pth',
  Model.vit_b: './models/sam_vit_b_01ec64.pth',
  Model.hq_vit_h: './models/sam_hq_vit_h.pth',
  Model.hq_vit_l: './models/sam_hq_vit_l.pth',
  Model.hq_vit_b: './models/sam_hq_vit_b.pth',
  Model.hq_vit_tiny: './models/sam_hq_vit_tiny.pth',
  Model.gd: './models/groundingdino_swint_ogc.pth',
}


def get_default_path_from_model(model: Model) -> str:
  path = DEFAULT_MODEL_LOCATION[model]
  if not path:
    raise ValueError(f'Invalid model: {model}')
  return path



from enum import Enum


# Image file formats with transparency
class OutputImageFormat(str, Enum):
  png = 'png'
  tiff = 'tiff'


# Some video codecs (and animated image formats) with alpha channel support that can be written out by FFmpeg
class OutputVideoCodec(str, Enum):
  prores = 'prores'
  vp9 = 'vp9'
  ffv1 = 'ffv1'
  apng = 'apng'
  gif = 'gif'


# Map video codecs to container format
codec_to_video_format = {
  OutputVideoCodec.prores: 'mov',
  OutputVideoCodec.vp9: 'webm',
  OutputVideoCodec.ffv1: 'mkv',
  OutputVideoCodec.apng: 'apng',
  OutputVideoCodec.gif: 'gif',
}


def get_video_fmt_from_codec(codec: OutputVideoCodec) -> str:
  format = codec_to_video_format[codec]
  if not format:
    raise ValueError(f'Invalid codec: {codec}')
  return format



# SPDX-License-Identifier: AGPL-3.0-only
#
# A pipeline to delete background from images or video.
# Selects the foreground using text prompts to GroundingDINO to detect objects,
#  and then segments everything using Segment-Anything or Segment-Anything HQ.
#
# Based on code in Grounded-Segment-Anything, copyright 2020 IDEA, Inc.
# ref: https://github.com/IDEA-Research/Grounded-Segment-Anything
#

import argparse
import numbers
import os
import sys

import torch
import groundingdino.util.inference as gd
# import segment_anything_hq as samhq # Noisy

from ezsam.lib.date import now
from ezsam.lib.downloader import download
from ezsam.lib.gpu import attempt_gpu_cleanup
from ezsam.models import Model, MODEL_URL, get_default_path_from_model
from ezsam.formats import OutputImageFormat, OutputVideoCodec
from ezsam.process import process_file
from ezsam.config.utils import cleanup_gdconfig_tmpfile, create_gdconfig_tmpfile
from ezsam.config.defaults import (
  DEFAULT_SAM_MODEL,
  DEFAULT_GROUNDING_DINO_CONFIG_PATH,
  DEFAULT_OUTPUT_DIR,
  DEFAULT_OUTPUT_SUFFIX,
  DEFAULT_BOX_THRESHOLD,
  DEFAULT_TEXT_THRESHOLD,
  DEFAULT_NMS_THRESHOLD,
  DEFAULT_IMAGE_FORMAT,
  DEFAULT_VIDEO_CODEC,
)


def parse_args(argv=None):
  def unit_interval(value: str):
    # First, try to parse as a number
    num = float(value)
    # Enforce that value is a real number between 0 and 1 inclusive.
    # Note boolean is subclass of real numbers.
    if isinstance(num, bool) or not isinstance(num, numbers.Real) or num < 0 or num > 1:
      raise ValueError(f'Error value {num} should be a float between 0 and 1 inclusive')
    return num

  parser = argparse.ArgumentParser('ezsam', add_help=True)
  # fmt: off
  parser.add_argument('input', nargs='+', help='Input image(s) or video(s) to process')
  parser.add_argument('-d', '--debug', action='store_true', help='Debug mode: annotate output with detection boxes and masks instead of removing backgrounds')
  parser.add_argument('--hq', '--use_sam_hq', action='store_true', help='Use SAM-HQ for object segmenting instead of SAM')
  parser.add_argument('--bmin', '--box_threshold', type=unit_interval, default=DEFAULT_BOX_THRESHOLD, help='Confidence threshold for object detection boxes [0,1]')
  parser.add_argument('--tmin', '--text_threshold', type=unit_interval, default=DEFAULT_TEXT_THRESHOLD, help='Confidence threshold for text prompts to be used at all [0,1]')
  parser.add_argument('--nmin', '--nms_threshold', type=unit_interval, default=DEFAULT_NMS_THRESHOLD, help='Threshold to remove lower quality object boxes during post processing [0,1]')
  parser.add_argument('--gd', '--gd_checkpoint', type=str, required=False, help='Path to GroundingDINO checkpoint file')
  parser.add_argument('-c', '--gconf', '--gd_config', type=str, required=False, help='Path to GroundingDINO config file')
  parser.add_argument('-m', '--sam_model', '--model', choices=['vit_h', 'vit_l', 'vit_b', 'vit_tiny'], required=False, help='SAM ViT version (vit_h/l/b). If omitted, will guess from checkpoint filename.')
  parser.add_argument('--sam', '--sam_checkpoint', type=str, required=False, help='Path to Segment-Anything checkpoint file')
  parser.add_argument('-p', '--prompts', '--prompt_string', nargs='*', help='Comma delimited list of prompts to use in foreground selection')
  parser.add_argument('-n', '--nprompts', '--nprompt_string', nargs='*', help='Comma delimited list of negative prompts to exclude from selection')
  parser.add_argument('--pfile', '--prompt_file', type=str, required=False, help='Path to file with foreground selection prompts, one per line')
  parser.add_argument('--npfile', '--nprompt_file', type=str, required=False, help='Path to file with negative prompts, one per line')
  parser.add_argument('--img', '--img_fmt', choices=[c.value for c in OutputImageFormat], default=DEFAULT_IMAGE_FORMAT, help='Image file format to use for output files(s)')
  parser.add_argument('--codec', '--vc', '--vcodec', choices=[c.value for c in OutputVideoCodec], default=DEFAULT_VIDEO_CODEC, help='Video codec to use for output file(s)')
  parser.add_argument('--nf', '--num_frames', type=int, required=False, help='Number of frames to process for each input video, for testing purposes')
  parser.add_argument('-s', '--output_suffix', type=str, default=DEFAULT_OUTPUT_SUFFIX, help='Suffix to append to processed output name(s) i.e. for ".out", src.jpg -> src.out.png')
  parser.add_argument('-o', '--output_dir', type=str, default=DEFAULT_OUTPUT_DIR, help='Directory to write processed output to')
  parser.add_argument('-k', '--keep', action='store_true', help='Keep temporary image files generated when processing video')
  parser.add_argument('--smem', '--show_memory', action='store_true', help='Show PyTorch CUDA memory summary on completion')
  # fmt: on
  return parser.parse_args(argv)


def main(argv=None):
  args = parse_args(argv)
  INPUT: list[str] = args.input or []
  DEBUG: bool = args.debug
  GD_CONFIG_PATH = args.gconf or DEFAULT_GROUNDING_DINO_CONFIG_PATH
  SAM_MODEL: str = args.sam_model or DEFAULT_SAM_MODEL
  USE_SAM_HQ: bool = args.hq
  model_name = ''
  if USE_SAM_HQ:
    model_name = 'hq_'
  elif SAM_MODEL == 'vit_tiny':
    raise ValueError('Must use vit_tiny with SAM-HQ only! Please try again with --hq if this is intended')
  model_name += SAM_MODEL
  default_sam_checkpoint_path: str = get_default_path_from_model(model_name)
  sam_checkpoint_path = args.sam or default_sam_checkpoint_path
  sam_checkpoint_specified: bool = not not args.sam
  default_gd_checkpoint_path: str = get_default_path_from_model(Model.gd)
  gd_checkpoint_path = args.gd or default_gd_checkpoint_path
  gd_checkpoint_specified: bool = not not args.gd
  print(f'args.bmin: {args.bmin}')
  BOX_THRESHOLD: float = args.bmin
  TEXT_THRESHOLD: float = args.tmin
  NMS_THRESHOLD: float = args.nmin
  PROMPT_STRING: str = ' '.join(args.prompts) if args.prompts else None
  NPROMPT_STRING: str = ' '.join(args.nprompts) if args.nprompts else None
  PROMPT_FILE: str = args.pfile
  NPROMPT_FILE: str = args.npfile
  IMG_FMT: OutputImageFormat = args.img or DEFAULT_IMAGE_FORMAT
  CODEC: OutputVideoCodec = args.codec or DEFAULT_VIDEO_CODEC
  NUM_TEST_FRAMES: int = args.nf
  OUTPUT_DIR: str = args.output_dir.rstrip('/')
  OUTPUT_SUFFIX: str = args.output_suffix
  CLEANUP: bool = not args.keep
  SHOW_MEMORY_SUMMARY: bool = args.smem
  print('---------------------')
  print('Running with options:')
  print(f'--input: {INPUT}')
  print(f'--debug: {DEBUG}')
  print(f'--box_threshold: {BOX_THRESHOLD}')
  print(f'--text_threshold: {TEXT_THRESHOLD}')
  print(f'--nms_threshold: {NMS_THRESHOLD}')
  print(f'--gd_checkpoint: {gd_checkpoint_path}')
  print(f'--gd_config: {GD_CONFIG_PATH}')
  print(f'--sam_model: {SAM_MODEL}')
  print(f'--sam_checkpoint: {sam_checkpoint_path}')
  print(f'--use_sam_hq: {USE_SAM_HQ}')
  print(f'--img_fmt: {IMG_FMT}')
  print(f'--vcodec: {CODEC}')
  print(f'--num_frames: {NUM_TEST_FRAMES}')
  print(f'--output_dir: {OUTPUT_DIR}')
  print(f'--output_suffix: {OUTPUT_SUFFIX}')
  print(f'--prompt_string: {PROMPT_STRING}')
  print(f'--nprompt_string: {NPROMPT_STRING}')
  print(f'--prompt_file: {PROMPT_FILE}')
  print(f'--nprompt_file: {NPROMPT_FILE}')
  print(f'--show_memory: {SHOW_MEMORY_SUMMARY}')
  print('---------------------')

  # Make a list of all user prompts for selecting foreground elements, prediction classes used in GroundingDINO

  def prompts_from_file(prompt_file: str) -> list[str]:
    res = []
    if prompt_file:
      with open(prompt_file, 'r') as f:
        res = [line.strip() for line in f.readlines()]
    return res
  
  def prompts_from_string(prompt_string: str):
    return [p.strip() for p in prompt_string.split(',')] if prompt_string else []
  
  def prompts_from_file_and_string(prompt_file: str, prompt_string: str) -> list[str]:
    file_prompts = prompts_from_file(prompt_file)
    string_prompts = prompts_from_string(prompt_string)
    all_prompts = file_prompts + string_prompts
    return all_prompts

  prompts = prompts_from_file_and_string(prompt_file=PROMPT_FILE, prompt_string=PROMPT_STRING)
  neg_prompts = prompts_from_file_and_string(prompt_file=NPROMPT_FILE, prompt_string=NPROMPT_STRING)
  if not prompts or len(prompts) <= 0:
    raise ValueError('You need to specify --prompts for selecting the foreground. See --help')
  print(f'Foreground selection prompts are: {prompts}')
  print(f'Negative (inverse) selection prompts are: {neg_prompts}')

  # Create output directory if it doesn't exist already
  if os.path.exists(OUTPUT_DIR):
    if not os.path.isdir(OUTPUT_DIR):
      raise ValueError(f'--output_dir is not a directory: {OUTPUT_DIR}. Please specify a directory to write output to!')
  else:
    print(f'Creating output directory: {OUTPUT_DIR} ...')
    os.makedirs(OUTPUT_DIR)

  # If GroundingDINO config file doesn't exist already, create temporary config under a different path
  using_gdconfig_tmpfile = False
  if not os.path.isfile(GD_CONFIG_PATH):
    print(f'Warning: No GroundingDINO config at: {GD_CONFIG_PATH}')
    GD_CONFIG_PATH = create_gdconfig_tmpfile()
    using_gdconfig_tmpfile = True

  # If checkpoint files are not specified, then download default checkpoints as needed
  print('Checking if models need to be downloaded ...')
  something_downloaded = False
  if not gd_checkpoint_specified and not os.path.isfile(default_gd_checkpoint_path):
    print('Downloading default GroundingDINO model ...')
    outdir = os.path.dirname(default_gd_checkpoint_path)
    download(MODEL_URL[Model.gd], outdir)
    something_downloaded = True
  if not sam_checkpoint_specified and not os.path.isfile(default_sam_checkpoint_path):
    outdir = os.path.dirname(default_sam_checkpoint_path)
    print('Downloading model ...')
    download(MODEL_URL[model_name], outdir)
    something_downloaded = True
  for checkpoint in [gd_checkpoint_path, sam_checkpoint_path]:
    if not os.path.isfile(checkpoint):
      raise Exception(f'Could not find model checkpoint file: {checkpoint}')
  if something_downloaded:
    print('Downloads finished')
  else:
    print('... no')

  if DEBUG:
    print('Debug mode active: output images will have bounding box and masks overlaying original')

  grounding_dino_model = None
  sam = None
  sam_predictor = None
  try:
    DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print(f'Running on: {DEVICE}')

    # Only running inference, not training models, so disable gradient calculation to reduce memory usage
    # ref: https://pytorch.org/docs/stable/generated/torch.no_grad.html
    with torch.no_grad():
      attempt_gpu_cleanup()

      print(f'{now()}: Loading GroundingDINO model ...')
      grounding_dino_model = gd.Model(model_config_path=GD_CONFIG_PATH, model_checkpoint_path=gd_checkpoint_path, device=DEVICE)

      print(f'{now()}: Loading SAM model and predictor ...')
      import segment_anything_hq as samhq

      sam = samhq.sam_model_registry[SAM_MODEL](checkpoint=sam_checkpoint_path)
      sam.to(device=DEVICE)
      sam_predictor = samhq.SamPredictor(sam)

      had_error = False
      for src in INPUT:
        try:
          process_file_args = {
            'src': src,
            'prompts': prompts,
            'neg_prompts': neg_prompts,
            'box_threshold': BOX_THRESHOLD,
            'text_threshold': TEXT_THRESHOLD,
            'nms_threshold': NMS_THRESHOLD,
            'sam_predictor': sam_predictor,
            'grounding_dino_model': grounding_dino_model,
            'img_fmt': IMG_FMT,
            'codec': CODEC,
            'num_test_frames': NUM_TEST_FRAMES,
            'output_suffix': OUTPUT_SUFFIX,
            'output_dir': OUTPUT_DIR,
            'debug': DEBUG,
            'cleanup': CLEANUP,
          }
          process_file(**process_file_args)
        except Exception as err:
          print(f'Error processing file {src}: {err}')
          had_error = True
      print(f'Finished all processing jobs at: {now()}')
  except Exception as err:
    print(err)
  finally:
    if using_gdconfig_tmpfile:
      cleanup_gdconfig_tmpfile()
    del grounding_dino_model
    del sam_predictor
    del sam
    attempt_gpu_cleanup()
    if torch.cuda.is_available() and SHOW_MEMORY_SUMMARY:
      print(torch.cuda.memory_summary())
    if had_error:
      print('Warning: at least one file could not be processed. Check for errors!')


if __name__ == '__main__':
  main(sys.argv[1:])



import os
from enum import Enum

import filetype


class InputMode(Enum):
  image = 1
  video = 2
  other = 3


def get_input_mode(file):
  # Check whether we are parsing a folder or file.
  # If file, check if it's an image or video using file signature (not extension).
  # ref: https://github.com/h2non/filetype.py
  input_mode = None
  if os.path.isdir(file):
    raise ValueError(f'Input is a directory, specify individual file(s) instead of: {file}')
  if os.path.isfile(file):
    kind = filetype.guess(file)
    if kind is None or kind.mime is None:
      raise ValueError(f'Could not determine input file type: {file}')
    if kind.mime.startswith('image'):
      # Workaround: treat all input GIFs as video, animated GIFs are much more important to support
      #  and OpenCV image read doesn't support still image GIF anyways
      if kind.mime == 'image/gif':
        print(f'Treating file as video GIF: {file}')
        input_mode = InputMode.video
      else:
        input_mode = InputMode.image
    elif kind.mime.startswith('video'):
      input_mode = InputMode.video
    else:
      input_mode = InputMode.other
  else:
    raise ValueError(f'Input is not a file: {file}')
  if input_mode == InputMode.other:
    raise ValueError(f'Can only process videos and images, skipping input: {file}')
  return input_mode
